#include <math.h>
#include "erlang.h"

/************************************************************

The Qlib program library has resulted from the work conducted 
over several years jointly by 

 Laboratory of Telecommunications Technology of 
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

double
Aerl(double x, double b)
{
  double a,da,e,s,prec;
  double relprec = 0.00000000001;
  double sqrtx = sqrt(x);
  double c = sqrt(2 / (M_PI * x));
  
  if (!(x > 0 && b > 0 && b < 1))
    return 0;
  
  if (b >= c)
    a = x + (b - c) / (c * c);
  else {
    s = sqrt(-log((2 - b/c) * b/c) * 2) - sqrtx * 2;
    a = (s*s+1)/4;
  }
  
  if ((e = Berl_d(x,a)) <= b) {
    a = x / (1 - b);
    e = Berl_d(x,a);
  } 
  prec = relprec*b;
  
  
  while (!(fabs(b-e) < prec)) {
    da = (b - e) * a / (((e - 1) * a + x + 1) * e - b) ;
    a = a + da;
    e = Berl_d(x,a);
  }
  
  
  return a;
}


    

