#include <math.h>
#include "queuel.h"
#include <stdlib.h>
#include <stdio.h>

/************************************************************

The Qlib program library has resulted from the work conducted
over several years jointly by

 Laboratory of Telecommunications Technology of
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

#define HMAX 10

int
sl_gauss_pivmax(double *a[], double *b, double *c, int n)
{
  /* Gauss-Jordan elimination method. Return value is 0 if n*n matrix a is
     singular - otherwise 1. c is the resulting vector. */
  
  int i, j, k, l, err;
  double max, pivot, coef, s;
  double **t;
  
  t=(double **) malloc((n+2) * sizeof(double *));
  for (i=0; i < (n+2); i++) {
    t[i]=(double *) malloc((n+2) * sizeof(double));
  }
  
  for(i=1; i<=n; i++) {
    for(j=1; j<=n; j++)
      t[i][j]=a[i-1][j-1];
    t[i][n+1]=b[i-1];
  }
  err=1;
  k=1;
  
  while (err==1 && k<n) {
    max=fabs(t[k][k]);
    l=k;
    for(i=k+1; i<=n; i++)
      if(max<fabs(t[i][k])) {
	max = fabs(t[i][k]);
	l = i;
      }
    if(max!=0) {
      if(l!=k)
	for(j=k; j<=n+1; j++) {
	  pivot=t[k][j];
	  t[k][j]=t[l][j];
	  t[l][j]=pivot;
	}
      pivot=t[k][k];
      for(i=k+1;i<=n;i++) {
	coef=t[i][k]/pivot;
	for(j=k+1;j<=n+1;j++)
	  t[i][j] -= coef*t[k][j];
      }
    }
    else
      err=0;
    k++;
  }
  
  if(t[n][n]==0)
    err=0;
  if(err==1) {
    c[n-1]=t[n][n+1]/t[n][n];
    for(i=n-1;i>=1;i--) {
      s=t[i][n+1];
      for(j=i+1;j<=n;j++)
	s-=t[i][j]*c[j-1];
      c[i-1]=s/t[i][i];
    }
  }
  
  for(i=0; i<(n+2); i++)
    free (t[i]);
  free(t);
  
  return(err);
}

double
Fend1(double x, int n, double rho)
{
  int m=floor(x);
  double lambda=0;
  double beta=0;
  int h=0;
  int i=0;
  int j=0;
  int k=0;
  double **a;
  double *P0;
  int err=0;
  double **f;
  double **F;
  double *pi;
  double ans=0;
  
  lambda = n*rho;
  beta = lambda*exp(-lambda);
  
  a = (double **) malloc((m*n+n) * sizeof(double *));
  for (i = 0; i <= (m*n+n-1); i++) {
    a[i] = (double *) malloc((m*n+n) * sizeof(double));
  }

  P0 = (double *) malloc(n * sizeof(double));
  
  f = (double **) malloc((HMAX*n+n) * sizeof(double *));
  for (i = 0; i <= (HMAX*n+n-1); i++) {
    f[i] = (double *) malloc((HMAX*n+n) * sizeof(double));
  }
  
  F = (double **) malloc(n * sizeof(double *));
  for (i = 0; i <= (n-1); i++) {
    F[i] = (double *) malloc(n * sizeof(double));
  }
  
  pi = (double *) malloc(n * sizeof(double));
  
  /* Initial values */
  
  for (i = 0; i <= (n-1); i++) {
    pi[i] = (double) 1/n;
  }
  
  for (k = 0; k <= (n-1); k++) {
    h = 0;
    for (i = 0; i <= (n-1); i++) {
      if (i==k)
	f[i][0] = 1;
      else
	f[i][0] = 0;
      for (j = 1; j <= i; j++) {
	f[i][j] = -lambda/((double) j) * f[i-1][j-1];
      }
    }
    while (h < HMAX) {
      h++;
      for (i = h*n; i <= (h*n + n - 1); i++) {
	f[i][0] = 0;
	for (j = 0; j <= (i-n); j++) {
	  f[i][0] = f[i][0] + f[i-n][j];
	}
	for (j = 1; j <= i; j++) {
	  if ( i%n == 0)
	    f[i][j] = -beta/((double) j) * f[i-1][j-1];
	  else
	    f[i][j] = -lambda/((double) j) * f[i-1][j-1];
	}
      }
    }
    for (i = 1; i <= n; i++) {
      F[i-1][k] = exp(lambda*h)*f[h*n + i - 1][0];
    }
        
  } /* End of k-loop */
  
  for(i=0; i<=(HMAX*n+n-1); i++)
    free (f[i]);
  free(f);

  err = sl_gauss_pivmax(F, pi, P0, n);
  
  for(i=0; i<=(n-1); i++)
    free (F[i]);
  free(F);
  free (pi);
  
  if (err == 0)
    return 0;
  
/*for (i = 0; i <= (n-1); i++) {
    printf("%f\n", P0[i]);
  }*/

  /* End of initial values */
  
  for (i = 1; i <= (m*n + n - 1); i++) {
    for (j = 0; j <= i; j++) {
      a[i][j] = 0;
    }
  }
  
  a[0][0] = P0[0];
  for (i = 1; i <= (n - 1); i++) {
    a[i][0] = P0[i];
    for (j = 1; j <= i; j++) {
      a[i][j] = -lambda/((double) j)*a[i-1][j-1];
    }
  }
  
  free (P0);
  
  for (i = n; i <= (m*n + n - 1); i++) {
    for (j = 0; j <= (i-n); j++) {
      a[i][0] = a[i][0] + a[i-n][j];
    }
    for (j = 1; j <= i; j++) {
      if ( i%n == 0)
	a[i][j] = -beta/((double) j)*a[i-1][j-1];
      else
	a[i][j] = -lambda/((double) j)*a[i-1][j-1];
    }
  }
  
  if ((x-m) == 0)
    ans = n*exp(lambda*x) * a[m*n + n - 1][0];
  else {
    for (j = 0; j  <= (m*n + n - 1); j++) {
      ans = ans +  a[m*n + n - 1][j] * pow((x-m),j);
    };
    ans = n*exp(lambda*x) * ans;
  }
  
  for(i=0; i<=(m*n+n-1); i++)
    free (a[i]);
  free(a);
  
  return ans;
}
