/* File: indexsort.c
 *
 * Description: File indexsort.c is a code file that contains implementations
 * for functions with which it's possible to sort data without chancing the
 * positions in the orginal data. 
 *
 * Author: Vesa Timonen <Vesa.Timonen@hut.fi>
 * Last modified: 13.1.2002
 */

#include "indexsort.h"

/* static variables that are used to determine the data structure whose data
   is compared and the function that returns a element from the stucture
   from pointed position. 
*/
static double (*getElem)(void *, int);
static void *dataPtr;


/* static comparison function that makes the comparison of previously defined
   structure elements using previously defined function.
*/
static int compare(const void *a, const void *b) {

  double t;

  t = getElem(dataPtr, *(int*)a) - getElem(dataPtr, *(int*)b);
  
  if (t > 0) 
    return 1;
  else if (t < 0) 
    return -1;
  
  return 0;
}


int (*getCompareFunc(void *dataPointer, double (*getElemFunc)(void *, int)))
     (const void *a, const void *b) {

  getElem = getElemFunc;
  dataPtr = dataPointer;
  return compare;
} 

void index_sort(int *base, int nbr_elems, void *data, 
	       double (*getElemFunc)(void *, int)) {

  qsort(base, nbr_elems, sizeof(int), getCompareFunc(data, getElemFunc));
}
