#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "queuel.h"

/************************************************************

The Qlib program library has resulted from the work conducted
over several years jointly by

 Laboratory of Telecommunications Technology of
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

static double logfactorial(unsigned long n1, unsigned long n);
static void make_coeff_vector(double *, int, double, double);

double Qmd1series(double x, double rho)
{
    double sum = 0, logrho, prev = 0, lf, fa = 1;
    long n;
  
    lf = logfactorial(1, x);
    logrho = log(rho);
  
    for (n = floor(x) + 1; n < x + 200000L; n++) {
        fa *= n;        
        if (fa > 1e50) {
            lf += log(fa);
            fa = 1;
        }
        sum += exp(n*(logrho + log(n-x) - rho) - lf) / fa; 
        if (n % 50 == 0) 
        {
            if (prev == sum)
                break;
            else
                prev = sum;
        }
    }
    return (sum*exp(rho*x)*(1-rho));
}


double Qmd1(double x, double rho)
{
  if (rho < .3) 
  {
      if (x < (9 + 15 * log10(.3 / rho)))
          return Qmd1series(x, rho);
      else
          return Qmd1approx(x, rho);
  }
  else
      if (x > 8)
          return Qmd1approx(x, rho);
      else
          return Qmd1poly(x, rho);
}

double 
Qmd1approx(double x, double rho)
{
  double r0 = (3 - rho) * (1 - rho) * log10(10/rho), er0;
  
  do
    {
      er0 = exp(r0);
      if (fabs(rho * (er0 - 1) - r0) < 0.000000001)
	break;
      r0 = r0 - ((rho * (er0 - 1) - r0) / (rho * er0 - 1));
    } 
  while (1);
  
  
  return((1 - rho)/(rho*er0 - 1) * exp(-r0 * x));
  
}

double 
Qmd1poly(double x, double rho)
{
  double alpha = rho * exp(-rho);
  double *P;
  double G = 0, floortemp, fracx;
  double t;
  int i;
  int floorx;
  
  fracx = modf(x, &floortemp);
  floorx = floortemp;
  
  P = calloc(floorx + 1, sizeof(double));
  make_coeff_vector(P, floorx, rho, alpha);
  
  if (fracx == 0)
    G = P[0];
  else
    for (i = 0; i <= floorx; i++) 
      G += P[i] * pow(fracx, i);
  
  free(P);
  t = exp(rho*x);
  
  return 1-G*t;
  
}

static void
make_coeff_vector(double *P, int floorx, double rho, double alpha)
{
  double prevsum;
  int i, j;
  
  for (i = 0; i <= floorx; i++)
    P[i] = 0;
  
  P[0] = 1 - rho;
  
  for (i = 1; i <= floorx; i++) 
    {
      for (j = 0, prevsum = 0; j < i; j++)
	prevsum += P[j];
      
      for (j = i; j > 0; j--)
	P[j] = -alpha * P[j-1] / j;
      
      P[0] = prevsum;
    }
  
}





/* Logarithm factorial. logfactorial(n1, n) = log(n1 * (n1+1) * ... * n) */

static double 
logfactorial(unsigned long n1, unsigned long n)
{
  double sum = 0;
  double mul = 1;
  unsigned long i;
  for (i = n1; i <= n; i++) {
    mul *= (double) i;
    if ((mul * (double)(i + 1)) > 1e250) {
      sum += log(mul);
      mul =1;
    }
  }
  sum += log(mul);
  return sum;
}








