#include <math.h>
#include "queuel.h"
#include <stdlib.h>

/************************************************************

The Qlib program library has resulted from the work conducted
over several years jointly by

 Laboratory of Telecommunications Technology of
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

double
Fekdn(double x, double rho, int k, int n)
{ 
  return Fmdn(x, rho, k*n);
}


double
IntFmdn(int N, double rho, int n)
{
  int i=0;
  double sum=0;
  double *probabilities;
  
  probabilities = Mdn(N, rho, n);
  for (i=0; i <= floor(n*(N+1)-1); i++)
    sum = sum + probabilities[i];
  return sum;  
}


double     
Fmdn(double x, double rho, int n)
{
  int i=0;
  int j=0;
  int v=0;
  double sum1=0;
  double sum2=0;
  double sum3=0;
  double result=0;
  double *probabilities;

  probabilities = Mdn((n-1), rho, n);

  if (x==0) {
    for (i=0; i <= (n-1); i++)
      sum1 = (sum1 + probabilities[i]);
    return sum1;
  }
  else {
    for (i=0; i <= (n-1); i++) {
      sum2=0;
      for (j=0; j <= i; j++)
        sum2 = (sum2 + probabilities[j]);
      sum3=0;
      for (v=0; v <= floor(x); v++)
        sum3 = ( sum3 + (pow((rho*n)*(v-x),(v*n+n-1-i))/factorial(v*n+n-1-i) * exp(-(rho*n)*(v-x))) );
      result = (result + (sum2*sum3));
    }
    return result;
  }
  
}


double				       
*Mdn(double x, double rho, int n)
{
  int i=0;
  int j=0;
  int k=0;
  int I=0;
  double *GuessP;
  double *P;
  double S=0;
  double maxi=0;
  double sum=0;
  double sum1=0;
  double sum2=0;
  double Difference=0;
  
  GuessP = (double *) malloc((1000)*sizeof(double));
  P = (double *) malloc((1000)*sizeof(double));

  for (i=0; i <= (n-1); i++)
    sum = ( sum + pow((rho*n),i)/factorial(i) );
   
  GuessP[0] = pow(( sum + pow((rho*n),n)/factorial(n) * 1/(1-(rho*n)/n) ), (-1));
  
  for(i=1; i < n; i++)
    GuessP[i] = ( GuessP[i-1] * (rho*n)/i  );

  i = (n - 1);
    
  while( GuessP[i] > (pow(10,-12)) ) {
    i = (i + 1);
    GuessP[i] = ( GuessP[i-1] * (rho*n)/n );
  }
     
  I = i; 

  for(i=(I+1); i<=(I+n); i++)
    GuessP[i] = ( GuessP[i-1] * (rho*n)/n );
    
  while( (maxi > (pow(10,-12))) || (maxi == 0) ) {
    maxi=0;
    sum1=0;
    
    for (j=0; j<=n; j++)
      sum1 = (sum1 + GuessP[j]);
    
    for(i=0; i <= I; i++) {
      sum2=0;
      
      for (k=(n+1);  k <= (n+i); k++) {
        sum2 = ( sum2 + GuessP[k]*pow((rho*n),(n+i-k))/factorial(n+i-k)*exp(-rho*n) );
      }
      
      P[i] = ( sum1 * pow((rho*n),i)/factorial(i) * exp(-rho*n) + sum2); 
    }
    
    for(i=(I+1); i <= (I+n); i++)
      P[i] = ( P[i-1] * (rho*n)/n );
    
    S=0;
    
    for(k=0; k <= (I+n); k++)
      S = (S + P[k]);
    
    for(i=0; i <= (I+n); i++)
      P[i] = P[i]/S;
    
    for(i=0; i <= (I+n); i++) {
      if ((P[i] - GuessP[i]) > 0)
        Difference = (P[i] - GuessP[i]);
      else
        Difference = -(P[i] - GuessP[i]); 
      if (Difference > maxi)
        maxi = Difference;
      GuessP[i] = P[i]; 
    }
    
  }  
  
  return P;
}






