/* File: network_data.c
 *
 * Description: File network_data.c is a code file that contains 
 * implementations for TData structure and functions for handling it.
 *
 * Author: Vesa Timonen <Vesa.Timonen@hut.fi>
 * Last modified: 14.2.2002
 */

#include "network_data.h"

TData init_Data(TData data, int maximum, int step) {

  int i;

  data = (TData) malloc(sizeof(*data));
  data->connections = (int**) malloc(maximum * sizeof(int*));
  data->distances = (double*) malloc(maximum * sizeof(double));
  for(i = 0; i < maximum; i++)
    data->connections[i] = (int*) malloc(2 * sizeof(int));
  data->max_nbr = maximum;
  data->nbr_connections = 0;
  data->step_size = step;
  
  return data;
}

void free_Data(TData data) {

  int i;

  for(i = 0; i < get_max(data); i++)
    free(data->connections[i]);
  free(data->connections);
  free(data->distances);
}

void set_Data(TData data, int node1, int node2, double dist) {

  int i;
  int curr_pos = get_nbr(data);

  /* if the max. size is already reached, the size has to be increased
     and new memory allocated */

  if (curr_pos == get_max(data)) {

    data->connections = 
      (int**) realloc(data->connections, 
		      (get_step(data) + get_max(data)) * sizeof(int*));
    data->distances = 
      (double*) realloc(data->distances, 
		      (get_step(data) + get_max(data)) * sizeof(double));
    for(i = 0; i < get_step(data); i++)
      data->connections[get_max(data) + i] = (int*) malloc(2 * sizeof(int));

    data->max_nbr += get_step(data);
  }

  data->connections[curr_pos][0] = node1;
  data->connections[curr_pos][1] = node2;
  data->distances[curr_pos] = dist;
  data->nbr_connections += 1;
}

void optim_Data(TData data) {

  if (get_nbr(data) < get_max(data)) {

    data->connections =
      (int**) realloc(data->connections, get_nbr(data) * sizeof(int*));

    data->distances =
      (double*) realloc(data->distances, get_nbr(data) * sizeof(double));

    data->max_nbr = get_nbr(data);
  }
}


int get_max(TData data) {
  return data->max_nbr;
}

int get_nbr(TData data) {
  return data->nbr_connections;
}

int get_step(TData data) {
  return data->step_size;
}

int get_n1(TData data, int pos) {
  return data->connections[pos][0];
}

int get_n2(TData data, int pos) {
  return data->connections[pos][1];
}

double get_dist(TData data, int pos) {
  return data->distances[pos];
}

double **Data2double(double **target, TData source, int nbr_nodes) {

  int i, j;

  for(i = 0; i < nbr_nodes; i++) {
    target[i][i] = 0;
    for(j = i + 1; j < nbr_nodes; j++) {
      target[i][j] = INF;
      target[j][i] = INF;
    }
  }

  for(i = 0; i < get_nbr(source); i++) {
    target[get_n1(source, i)][get_n2(source, i)] = get_dist(source, i);
    target[get_n2(source, i)][get_n1(source, i)] = get_dist(source, i);
  }

  return target;
}

TData double2Data(TData target, double **source, int nbr_nodes) {

  int i, j;

  for(i = 0; i < nbr_nodes; i++)
    for(j = i + 1; j < nbr_nodes; j++)
      if (source[i][j] != INF)
	set_Data(target, i, j, source[i][j]);

  optim_Data(target);
  return target;
}
