#include <math.h>
#include "queuel.h"
#include <stdlib.h>

/************************************************************

The Qlib program library has resulted from the work conducted
over several years jointly by

 Laboratory of Telecommunications Technology of
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

static double get_z(int, int, double, double *);

/*    Sigma D_i/D/1 queue length distribution  */
/*    D[]   -  List of periods for all sources */
/*    N     -  Number of sources               */


double
Qsdd1(double x, double *D, long N)
{
  double sum = 0, psi, z, s, t, rho = 0, iptr;
  double *p;
  int i, n, d;
  
  p = calloc((int)N, sizeof(double));
  
  for (i = 0; i < N; i++)
    rho += 1/D[i];
  
  for (n = x + 1; n <= (N - rho * x) / (1 - rho); n++) 
    {
      for(i=0, d=0; i < N; i++) 
	{
	  p[i] = modf((n-x)/D[i], &iptr);
	  d  += iptr;
	}
      
      /* if n - d >= N we can't find values for z */
      
      if (n - d >= N) continue;
      
      z = get_z(n, (int)N, d, p);
      if (z == 0) continue;
      
      for (i = 0, psi = 0, s = 0, t = 0; i < N; i++) 
	{
	  psi += log(1 - p[i] + p[i] * z);  
	  s += p[i] * (1 - p[i]) * z / sqr(1 - p[i] + p[i] * z);
	  t += 1 / D[i] / (1 - p[i] + p[i] * z);
	}
      
      sum += exp(psi - (n-d) * log(z)) / sqrt(2*M_PI * s) * (1 - t);
    }
  
  free(p);
  return sum;
}

/* This function returns the value of z(n) for a Sigma D_i/D/1 system */
/* z is approximated by a function */

static double 
get_z(int n, int N, double d, double *p)
{  
  int i;
  double fz = 0, z, df0 = 0;
  
  /* Calculate f'(0)  */
  
  for (i = 0; i < N; i++) 
    df0 += p[i]/(1-p[i]);
  
  if (df0 == 0) return 0;
  
  z = N / df0 * (n - d) / (N - (n - d));
  
  for (i = 0; i < N; i++) 
    fz += p[i] *  z / (1 - p[i] + p[i] * z);
  
  while (fabs(fz - n + d) > 0.001) 
    {
      z  = (n - d) / (df0 - (n-d)*(df0/fz - 1/z));
      for (i = 0, fz = 0; i < N; i++) 
	fz += p[i] *  z / (1 - p[i] + p[i] * z);
    }
  return z;
}  











