#include <math.h>
#include "erlang.h"

/************************************************************

The Qlib program library has resulted from the work conducted
over several years jointly by

 Laboratory of Telecommunications Technology of
 Helsinki University of Technology (HUT) and

 VTT Information Technology, Telecommunications

The following individuals have contributed to the library:

 Samuli Aalto, Jani Lakkakorpi, Ilkka Norros, Anssi Pirhonen
 and Jorma Virtamo.

The programs can be used and modified freely.

No claims are made about the correctness of the programs
and no liability is taken for any damage caused by the use
of the programs in the library.

Reports of bugs in the program can be sent to qlib@tct.hut.fi

*************************************************************/

static double secant_iter(double x0, double x1, double b, double a)
{
    /* 
       xdiag - diagonal x values from iteration
       xv    - vertical x values from normal secant method
       yv    - vertical y values from normal secant method
    */
  
    double xdiag[10], xv[10], yv[10], xt, xprev = 0;
    int i,j;
  
    xv[0] = xdiag[0] = x0;
    yv[0] = Berl_d(x0, a) - b;
  
    xv[1] = x1;
    yv[1] = Berl_d(x1, a) - b;
  
    xdiag[1] = x1 - yv[1] * (x1 - x0) / (yv[1] - yv[0]);
  
    for (i = 2; i < 10; i++) 
    {
        xv[i] = xt = xv[i-1] - yv[i-1] * (xv[i-1] - xv[0]) / (yv[i-1] - yv[0]);
        yv[i] = Berl_d(xt, a) - b;
        if (fabs(yv[i]) < 1e-15) 
            return xt;
      
        for (j = 0; j < i; j++) 
            xt = xt - yv[i] * (xt - xdiag[j]) / (yv[i] - yv[j]);
        xdiag[i] = xt;
        if (xt == xprev)
            return xt;
        else
            xprev = xt;
    }
    return xprev;
  
}


double Xerl(double b, double a)
{
    double sqrta, sqrt2pia, x, s, b2;
    double relprec, abserror, n;
  
    if (!(a > 0 && b > 0 && b < 1))
        return 0;
  
    sqrta = sqrt(a);
    sqrt2pia = sqrt(2*M_PI) * sqrta;
  
    abserror = 0.333;
    relprec = 0.000000001;
  
    if (b * sqrt2pia < 2) {
        x = sqrt(-log((4-b*sqrt2pia)*b*sqrt2pia/4)*2)+sqrta*2;
        x = (x*x-3)/4 - 1/(16*b + 4/9 + a);
        if (x*relprec > abserror)
            return x;
    }
  
    b2 = 1;
    s  = 1/a;
  
    for (x = s; b2 > b; x += s)
        b2 = b2 /(x + b2);
  
    x -= s;
    n  = x*a - 1;
  
    return(secant_iter(n, n+1, b, a));
  
}
