/* File: interface.tm
 * __
 * UpDates:
 * 18.04.2002 LnkCreateNet and LnkDijkstra added by Vesa
 * 05.05.2002 LnkDijkstra usage updated...
 * 27.05.2002 LnkCreateNet renamed as LnkGabrielNet and the interface&usage updated
 *            Function LnkShowNetwork added
 * 30.05.2002 Minor modifications to LnkGabrielNet/LnkShowNetwork: Context of optional arguments forced to 'System`'
 *            Now because context of opt.args in qlib.m is the same, no shadowing appears while simultaneous use
 */

#include "erlang.h"
#include "queuel.h"
#include "qlib.h"

:Evaluate:      LnkBerld::usage = "LnkBerld[x, a] returns the Erlang loss probability for a trunk group x, where x is a real valued variable (double). Second parameter is the Poisson traffic a."
:Begin:
:Function:      Berl_d
:Pattern:       LnkBerld[x_, a_]
:Arguments:     {x,a}
:ArgumentTypes: {Real, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkBerli::usage = "LnkBerli[n, a] returns the Erlang loss probability for a integer valued trunk group n and Poisson traffic a."
:Begin:
:Function:      Berl_i
:Pattern:       LnkBerli[n_, a_]
:Arguments:     {n,a}
:ArgumentTypes: {Integer, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkXerl::usage = "LnkXerl[b, a] returns the number of trunks for a loss probability b and Poisson traffic a."
:Begin:
:Function:      Xerl
:Pattern:       LnkXerl[b_, a_]
:Arguments:     {b,a}
:ArgumentTypes: {Real, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkAerl::usage = "LnkAerl[x, b] returns the Poisson traffic value for a trunk group x and loss probability b."
:Begin:
:Function:      Aerl
:Pattern:       LnkAerl[x_, b_]
:Arguments:     {x,b}
:ArgumentTypes: {Real, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkQndd1::usage = "LnkQndd1[x, N, D] returns the virtual waiting time distribution for the N*D/D/1 queuing system. The system has N deterministic sources with the same period D, so that the load level of the system is N/D."
:Begin:
:Function:      Qndd1
:Pattern:       LnkQndd1[x_, N_, D_]
:Arguments:     {x, N, D}
:ArgumentTypes: {Real, Integer, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkQsdd1::usage = "LnkQsdd1[x, D, N] returns the virtual waiting time distribution for the [Sigma]Di/D/1 queuing system. The system has N deterministic sources, each having it's own period D, and a constant service time. Table of periods is given by D."
:Begin:
:Function:      Qsdd1
:Pattern:       LnkQsdd1[x_, D:{__}]
:Arguments:     {x, D}
:ArgumentTypes: {Real, RealList}
:ReturnType:    Real
:End:

:Evaluate:      LnkQmd1::usage = "LnkQmd1[x, rho] returns the virtual waiting time distribution for the M/D/1 queuing system. The system has Poisson arrivals and deterministic service time. Rho is the load level of the system."
:Begin:
:Function:      Qmd1
:Pattern:       LnkQmd1[x_, rho_]
:Arguments:     {x, rho}
:ArgumentTypes: {Real, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkQmxd1::usage = "LnkQmxd1[x, rho, D] returns the unfinished work tail distribution for the M^x/D/1 queuing system. The system has Poisson batch arrivals and deterministic service time. Parameter x is the amount of unfinished work in the system, rho is the load level of the system and parameter D is the distribution (table) of the batch size probabilities {p1, p2, p3...}. The sum of these probabilities has to be exactly 1."
:Begin:
:Function:      Qmxd1
:Pattern:       LnkQmxd1[x_, rho_, D:{__}]
:Arguments:     {x, rho, D}
:ArgumentTypes: {Real, Real, RealList}
:ReturnType:    Real
:End:

:Evaluate:      LnkMg1::usage = "LnkMg1[x, rho, n] gives the queue length probabilities for the M/G/1 queuing model with load level rho and service time distribution function N ( If N=1: If[x<1,0,1]; If N=2: 1-Exp[-x] )."
:Begin:
:Function:      lnkMg1
:Pattern:       LnkMg1[x_, rho_, D_]
:Arguments:     {x, rho, D}
:ArgumentTypes: {Integer, Real, Manual}
:ReturnType:    Manual
:End:

:Evaluate:      LnkSumMd1::usage = "LnkSumMd1[x, rho] calculates the state probabilities of a M/D/1 queue with load level rho."
:Begin:
:Function:      lnkSumMd1
:Pattern:       LnkSumMd1[x_, rho_]
:Arguments:     {x, rho}
:ArgumentTypes: {Integer, Real}
:ReturnType:    Manual
:End:

:Evaluate:      LnkRecMd1::usage = "LnkRecMd1[x, rho] calculates the state probabilities of a M/D/1 queue with load level rho. This function uses a recursive algorithm."
:Begin:
:Function:      lnkRecMd1
:Pattern:       LnkRecMd1[x_, rho_]
:Arguments:     {x, rho}
:ArgumentTypes: {Integer, Real}
:ReturnType:    Manual
:End:

:Evaluate:      LnkIntFmd1::usage = "LnkIntFmd1[N, rho] is the M/D/1 (virtual) waiting time distribution function (for integral values of x) with load level rho. Service time h is assumed to be 1."
:Begin:
:Function:      IntFmd1
:Pattern:       LnkIntFmd1[N_, rho_]
:Arguments:     {N, rho}
:ArgumentTypes: {Integer, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkFmd1::usage = "LnkFmd1[x, rho] is the M/D/1 (virtual) waiting time distribution function (it calculates the probability that waiting time W is smaller than x) with load level rho. Service time h is assumed to be 1."
:Begin:
:Function:      Fmd1
:Pattern:       LnkFmd1[x_, rho_]
:Arguments:     {x, rho}
:ArgumentTypes: {Real, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkMdn::usage = "LnkMdn[x, rho, n] calculates the state probabilities of a M/D/n queue with load level rho and n servers. Accuracy goal is 10^-12."
:Begin:
:Function:      lnkMdn
:Pattern:       LnkMdn[x_, rho_, n_]
:Arguments:     {x, rho, n}
:ArgumentTypes: {Real, Real, Integer}
:ReturnType:    Manual
:End:

:Evaluate:      LnkIntFmdn::usage = "LnkIntFmdn[N, rho, n] is the M/D/n (virtual) waiting time distribution function (for integral values of x) with load level rho. Service time h is assumed to be 1."
:Begin:
:Function:      IntFmdn
:Pattern:       LnkIntFmdn[N_, rho_, n_]
:Arguments:     {N, rho, n}
:ArgumentTypes: {Integer, Real, Integer}
:ReturnType:    Real
:End:

:Evaluate:      LnkFmdn::usage = "LnkFmdn[x, rho, n] is the M/D/n (virtual) waiting time distribution function (it calculates the probability that waiting time W is smaller than x) with load level rho. Service time h is assumed to be 1."
:Begin:
:Function:      Fmdn
:Pattern:       LnkFmdn[x_, rho_, n_]
:Arguments:     {x, rho, n}
:ArgumentTypes: {Real, Real, Integer}
:ReturnType:    Real
:End:

:Evaluate:      LnkFekdn::usage = "LnkFekdn[x, rho, k, n] is the E_k/D/n (virtual) waiting time distribution function (it calculates the probability that waiting time W is smaller than x) with load level rho. Service time h is assumed to be 1."
:Begin:
:Function:      Fekdn
:Pattern:       LnkFekdn[x_, rho_, k_, n_]
:Arguments:     {x, rho, k, n}
:ArgumentTypes: {Real, Real, Integer, Integer}
:ReturnType:    Real
:End:

:Evaluate:      LnkFend1::usage = "LnkFend1[x, n, rho] is the E_n/D/1 (virtual) waiting time distribution function (it calculates the probability that waiting time W is smaller than x) with load level rho. Service time h is assumed to be 1."
:Begin:
:Function:      Fend1
:Pattern:       LnkFend1[x_, n_, rho_]
:Arguments:     {x, n, rho}
:ArgumentTypes: {Real, Integer, Real}
:ReturnType:    Real
:End:

:Evaluate:      LnkBmitra::usage = "LnkBmitra[j, J, C, d, a] returns the blocking probability for a traffic class j (1<=j<=J) from a number of classes J. C is the total capacity of the link. Each traffic class is characterized by it's bandwidth. d is the list of these values and a is the list of traffic intensities for each class."
:Begin:
:Function:      lnkbmitra
:Pattern:       LnkBmitra[j_Integer, J_Integer, C_, d:{__}, a:{__}]
:Arguments:     {j, J, C, d, a}
:ArgumentTypes: {Integer, Integer, Real, IntegerList, RealList}
:ReturnType:    Real
:End:

:Evaluate:      LnkBkaufman::usage = "LnkBkaufman[j, J, C, d, a] returns the blocking probability for a traffic class j (1<=j<=J) from a number of classes J. C is the total capacity of the link. Each traffic class is characterized by it's bandwidth. d is the list of these values and a is the list of traffic intensities for each class. LnkBkaufman is implemented by using the Kaufman and Roberts recursion formula."
:Begin:
:Function:      lnkbkaufman
:Pattern:       LnkBkaufman[k_Integer, K_Integer, C_, dk:{__}, ak:{__}]
:Arguments:     {k, K, C, dk, ak}
:ArgumentTypes: {Integer, Integer, Integer, IntegerList, RealList}
:ReturnType:    Real
:End:

/* begin: LnkGabrielNet */

:Begin:
:Function:       lnkgabrielnet
:Pattern:        LnkGabrielNetCore[NbrOfNodes_Integer, Scl_Real, Shp_Integer]
:Arguments:      { NbrOfNodes, Scl, Shp }
:ArgumentTypes:  { Integer, Real, Integer }
:ReturnType:     Manual
:End:

:Evaluate: Options[LnkGabrielNet] = 
	{System`Scale -> 1., System`Shape -> Circle, System`CreateGraph -> False, System`Index -> False, System`Graph -> False};

:Evaluate: LnkGabrielNet[nbrNds_, options___] := 
  Module[{d, points, graph = {}, i, j, scl, shp},
  scl = Scale /. {options} /. Options[LnkGabrielNet];
  shp = Shape /. {options} /. Options[LnkGabrielNet] /. {Circle -> 0, Box -> 1};
  {points, d} = LnkGabrielNetCore[nbrNds, N[scl], shp];
  d = d /. {-1. -> Infinity};
  If[(CreateGraph /. {options} /. Options[LnkGabrielNet]) || (Graph /. {options} /. Options[LnkGabrielNet]), (* Create the graph *)
     {d, points, RCore[{d, points},{},{Index, Graph} /. {options} /. Options[LnkGabrielNet]]},
    (* do not create or return the graph *)
     {d, points}]];
  
:Evaluate: LnkGabrielNet::usage = "LnkGabrielNet[NbrOfNodes] generates a random network containing 'NbrOfNodes' nodes using an external c-program via MathLink. The positions of the nodes are drawn from uniform distribution and node pairs satisfying the gabriel criterion are connected. Function returns a list containing two or three elements: distance/cost matrix, list of the node coordinates and, if wanted, a Graphics object containing an image of the network.\nOptional arguments:\n Shape      \tCircle/Box\t the shape of the network (default 'Circle')\n Scale      \t          \t the size of the circle radius/box edge (def. '1')\n CreateGraph\tFalse/True\t whether to create the graph (def. 'False')\n Index      \tFalse/True\t whether to include node indices to the graph (def. 'False')\n Graph      \tFalse/True\t whether to show the graph (def. 'False')"

/* end: LnkGabrielNet */

/* begin: LnkDijkstra */

:Begin:
:Function:       lnkdijkstra
:Pattern:        LnkDijkstraCore[Root_Integer, Target_Integer, Distances_List]
:Arguments:      { Root, Target, Distances }
:ArgumentTypes:  { Integer, Integer, Manual }
:ReturnType:     Manual
:End:

:Evaluate: LnkDijkstra[Distances_, Root_, Target_:0] := 
  LnkDijkstraCore[Root - 1, Target - 1, Distances /. {Infinity -> -1}] + {1, 1, 0};

:Evaluate: LnkDijkstra::usage = "LnkDijkstra[dist, root, dest] finds out the shortest path from node 'root' to node 'dest' with Dijkstras algorithm using an external c-program via MathLink. If 'dest' is not given, function finds out paths to all nodes in the net. Matrix 'dist' is the distance/cost matrix between all nodes in the network (if there is no connection, the distance/cost must be 'Infinity'). Function returns a list containing three elements: list of nodes that belong to the path, list of previous nodes of corresponding node in the first list and total distance/cost of corresponding node in the first list to the node 'root'. \nIf there is no connection to some nodes, those are left out from the results. If there is no connection from 'root' to specific 'dest', result will be list containing only the root node."

/* end: LnkDijkstra */

/* begin: LnkShowNetwork */

/* NOTE! This function is a pure Mathematica function that uses no external programs          */
/*       It is added to make possible to create graphs without installing the qlib.m package  */
/*       This implementation of ShowRoutes is identical with corresponding function in qlib.m */

:Evaluate: Options[LnkShowNetwork] = {System`Index -> False, System`Graph -> False};

:Evaluate: LnkShowNetwork[GraphData_, RouteData_, options___] := 
  RCore[GraphData, RouteData, {Index, Graph} /. {options} /. Options[LnkShowNetwork]];

:Evaluate: RCore[GraphData_, RouteData_, {Index_, Graph_}] := Module[{net={}, graph, root},
    If[Length[GraphData] == 2, (* no graph created *)
       Do[If[GraphData[[1]][[i,j]] != Infinity, (* points i and j connected *)
             net = net~Join~{{GraphData[[2,i]], GraphData[[2,j]]}}],
          {i,1,Length[GraphData[[1]]]},{j,i+1,Length[GraphData[[1]]]}];
       net = {GrayLevel[.3], Line /@ net};
       net = {net, GrayLevel[0], AbsolutePointSize[5], Point /@ GraphData[[2]]};
       If[Index, (* add indices *)
          net = {net, RGBColor[0, 0.4, 0.2],
          MapIndexed[Text[First[#2], #1, {-2, 0}, Background -> GrayLevel[.95]]&, GraphData[[2]]]}];
       net = Graphics[net],
       net = GraphData[[3]] (* else *)];
    
    If[RouteData=!={}, (* path info is known *)
      graph = Line[GraphData[[2, #]]] & /@ Transpose[Take[RouteData, 2]];
      root = {PointSize[.035], RGBColor[1, .5, .1], Point[GraphData[[2, RouteData[[1, 1]]]]]};
      graph = {root, {AbsoluteThickness[2], graph}};
      If[Index, graph = {graph, RGBColor[.3, .1, 0], 
         Text[#, GraphData[[2, #]], {-2, 0}, Background -> GrayLevel[.95]] & /@ RouteData[[1]]}];
      If[Length[GraphData]==3,graph = {GraphData[[3]], Graphics[graph]},graph = {net, Graphics[graph]}],
      graph=net (* else *)];
      
    If[Graph, Show[graph, AspectRatio -> Automatic, PlotRange -> All];];
    graph];

:Evaluate: LnkShowNetwork::usage = "LnkShowNetwork[netData, pathData] List 'netData' contains information of the network in the output form of function 'LnkGabrielNet' and list 'pathData' contains information of the route(s) in the output form of function 'LnkDijkstra'. Function returns a list of two Graphics objects that contains an image of the network and the path(s).\nIf list 'pathData' is empty i.e. {}, returns the function only the graph of the network.\nOptional arguments:\n Index \tFalse/True \t whether to include node indices to the graph (default 'False')\n Graph \tFalse/True \t whether to show the created graph (def. 'False')"

/* end: LnkShowNetwork */